==> opencryptoki_2.2.6%2Bdfsg.orig.tar.gz: opencryptoki-2.2.6+dfsg/usr/lib/pkcs11/common/mech_dsa.c <==
/*
             Common Public License Version 0.5

             THE ACCOMPANYING PROGRAM IS PROVIDED UNDER THE TERMS OF
             THIS COMMON PUBLIC LICENSE ("AGREEMENT"). ANY USE,
             REPRODUCTION OR DISTRIBUTION OF THE PROGRAM CONSTITUTES
             RECIPIENT'S ACCEPTANCE OF THIS AGREEMENT.

             1. DEFINITIONS

             "Contribution" means: 
                   a) in the case of the initial Contributor, the
                   initial code and documentation distributed under
                   this Agreement, and 

                   b) in the case of each subsequent Contributor:
                   i) changes to the Program, and
                   ii) additions to the Program;

                   where such changes and/or additions to the Program
                   originate from and are distributed by that
                   particular Contributor. A Contribution 'originates'
                   from a Contributor if it was added to the Program
                   by such Contributor itself or anyone acting on such
                   Contributor's behalf. Contributions do not include
                   additions to the Program which: (i) are separate
                   modules of software distributed in conjunction with
                   the Program under their own license agreement, and
                   (ii) are not derivative works of the Program.


             "Contributor" means any person or entity that distributes
             the Program.

             "Licensed Patents " mean patent claims licensable by a
             Contributor which are necessarily infringed by the use or
             sale of its Contribution alone or when combined with the
             Program. 

             "Program" means the Contributions distributed in
             accordance with this Agreement.

             "Recipient" means anyone who receives the Program under
             this Agreement, including all Contributors.

             2. GRANT OF RIGHTS

                   a) Subject to the terms of this Agreement, each
                   Contributor hereby grants Recipient a
                   non-exclusive, worldwide, royalty-free copyright
                   license to reproduce, prepare derivative works of,
                   publicly display, publicly perform, distribute and
                   sublicense the Contribution of such Contributor, if
                   any, and such derivative works, in source code and
                   object code form.

                   b) Subject to the terms of this Agreement, each
                   Contributor hereby grants Recipient a
                   non-exclusive, worldwide, royalty-free patent
                   license under Licensed Patents to make, use, sell,
                   offer to sell, import and otherwise transfer the
                   Contribution of such Contributor, if any, in source
                   code and object code form. This patent license
                   shall apply to the combination of the Contribution
                   and the Program if, at the time the Contribution is
                   added by the Contributor, such addition of the
                   Contribution causes such combination to be covered
                   by the Licensed Patents. The patent license shall
                   not apply to any other combinations which include
                   the Contribution. No hardware per se is licensed
                   hereunder.

                   c) Recipient understands that although each
                   Contributor grants the licenses to its
                   Contributions set forth herein, no assurances are
                   provided by any Contributor that the Program does
                   not infringe the patent or other intellectual
                   property rights of any other entity. Each
                   Contributor disclaims any liability to Recipient
                   for claims brought by any other entity based on
                   infringement of intellectual property rights or
                   otherwise. As a condition to exercising the rights
                   and licenses granted hereunder, each Recipient
                   hereby assumes sole responsibility to secure any
                   other intellectual property rights needed, if any.

                   For example, if a third party patent license is
                   required to allow Recipient to distribute the
                   Program, it is Recipient's responsibility to
                   acquire that license before distributing the
                   Program.

                   d) Each Contributor represents that to its
                   knowledge it has sufficient copyright rights in its
                   Contribution, if any, to grant the copyright
                   license set forth in this Agreement.

             3. REQUIREMENTS

             A Contributor may choose to distribute the Program in
             object code form under its own license agreement, provided
             that:
                   a) it complies with the terms and conditions of
                   this Agreement; and

                   b) its license agreement:
                   i) effectively disclaims on behalf of all
                   Contributors all warranties and conditions, express
                   and implied, including warranties or conditions of
                   title and non-infringement, and implied warranties
                   or conditions of merchantability and fitness for a
                   particular purpose;

                   ii) effectively excludes on behalf of all
                   Contributors all liability for damages, including
                   direct, indirect, special, incidental and
                   consequential damages, such as lost profits;

                   iii) states that any provisions which differ from
                   this Agreement are offered by that Contributor
                   alone and not by any other party; and

                   iv) states that source code for the Program is
                   available from such Contributor, and informs
                   licensees how to obtain it in a reasonable manner
                   on or through a medium customarily used for
                   software exchange.

             When the Program is made available in source code form:
                   a) it must be made available under this Agreement;
                   and
                   b) a copy of this Agreement must be included with
                   each copy of the Program. 

             Contributors may not remove or alter any copyright notices
             contained within the Program.

             Each Contributor must identify itself as the originator of
             its Contribution, if any, in a manner that reasonably
             allows subsequent Recipients to identify the originator of
             the Contribution. 


             4. COMMERCIAL DISTRIBUTION

             Commercial distributors of software may accept certain
             responsibilities with respect to end users, business
             partners and the like. While this license is intended to
             facilitate the commercial use of the Program, the
             Contributor who includes the Program in a commercial
             product offering should do so in a manner which does not
             create potential liability for other Contributors.
             Therefore, if a Contributor includes the Program in a
             commercial product offering, such Contributor ("Commercial
             Contributor") hereby agrees to defend and indemnify every
             other Contributor ("Indemnified Contributor") against any
             losses, damages and costs (collectively "Losses") arising
             from claims, lawsuits and other legal actions brought by a
             third party against the Indemnified Contributor to the
             extent caused by the acts or omissions of such Commercial
             Contributor in connection with its distribution of the
             Program in a commercial product offering. The obligations
             in this section do not apply to any claims or Losses
             relating to any actual or alleged intellectual property
             infringement. In order to qualify, an Indemnified
             Contributor must: a) promptly notify the Commercial
             Contributor in writing of such claim, and b) allow the
             Commercial Contributor to control, and cooperate with the
             Commercial Contributor in, the defense and any related
             settlement negotiations. The Indemnified Contributor may
             participate in any such claim at its own expense.


             For example, a Contributor might include the Program in a
             commercial product offering, Product X. That Contributor
             is then a Commercial Contributor. If that Commercial
             Contributor then makes performance claims, or offers
             warranties related to Product X, those performance claims
             and warranties are such Commercial Contributor's
             responsibility alone. Under this section, the Commercial
             Contributor would have to defend claims against the other
             Contributors related to those performance claims and
             warranties, and if a court requires any other Contributor
             to pay any damages as a result, the Commercial Contributor
             must pay those damages.


             5. NO WARRANTY

             EXCEPT AS EXPRESSLY SET FORTH IN THIS AGREEMENT, THE
             PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT
             WARRANTIES OR CONDITIONS OF ANY KIND, EITHER EXPRESS OR
             IMPLIED INCLUDING, WITHOUT LIMITATION, ANY WARRANTIES OR
             CONDITIONS OF TITLE, NON-INFRINGEMENT, MERCHANTABILITY OR
             FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is solely
             responsible for determining the appropriateness of using
             and distributing the Program and assumes all risks
             associated with its exercise of rights under this
             Agreement, including but not limited to the risks and
             costs of program errors, compliance with applicable laws,
             damage to or loss of data, programs or equipment, and
             unavailability or interruption of operations. 

             6. DISCLAIMER OF LIABILITY
             EXCEPT AS EXPRESSLY SET FORTH IN THIS AGREEMENT, NEITHER
             RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY
             FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
             OR CONSEQUENTIAL DAMAGES (INCLUDING WITHOUT LIMITATION
             LOST PROFITS), HOWEVER CAUSED AND ON ANY THEORY OF
             LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
             (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
             OF THE USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE
             OF ANY RIGHTS GRANTED HEREUNDER, EVEN IF ADVISED OF THE
             POSSIBILITY OF SUCH DAMAGES.

             7. GENERAL

             If any provision of this Agreement is invalid or
             unenforceable under applicable law, it shall not affect
             the validity or enforceability of the remainder of the
             terms of this Agreement, and without further action by the
             parties hereto, such provision shall be reformed to the
             minimum extent necessary to make such provision valid and
             enforceable.


             If Recipient institutes patent litigation against a
             Contributor with respect to a patent applicable to
             software (including a cross-claim or counterclaim in a
             lawsuit), then any patent licenses granted by that
             Contributor to such Recipient under this Agreement shall
             terminate as of the date such litigation is filed. In
             addition, If Recipient institutes patent litigation
             against any entity (including a cross-claim or
             counterclaim in a lawsuit) alleging that the Program
             itself (excluding combinations of the Program with other
             software or hardware) infringes such Recipient's
             patent(s), then such Recipient's rights granted under
             Section 2(b) shall terminate as of the date such
             litigation is filed.

             All Recipient's rights under this Agreement shall
             terminate if it fails to comply with any of the material
             terms or conditions of this Agreement and does not cure
             such failure in a reasonable period of time after becoming
             aware of such noncompliance. If all Recipient's rights
             under this Agreement terminate, Recipient agrees to cease
             use and distribution of the Program as soon as reasonably
             practicable. However, Recipient's obligations under this
             Agreement and any licenses granted by Recipient relating
             to the Program shall continue and survive. 

             Everyone is permitted to copy and distribute copies of
             this Agreement, but in order to avoid inconsistency the
             Agreement is copyrighted and may only be modified in the
             following manner. The Agreement Steward reserves the right
             to publish new versions (including revisions) of this
             Agreement from time to time. No one other than the
             Agreement Steward has the right to modify this Agreement.

             IBM is the initial Agreement Steward. IBM may assign the
             responsibility to serve as the Agreement Steward to a
             suitable separate entity. Each new version of the
             Agreement will be given a distinguishing version number.
             The Program (including Contributions) may always be
             distributed subject to the version of the Agreement under
             which it was received. In addition, after a new version of
             the Agreement is published, Contributor may elect to
             distribute the Program (including its Contributions) under
             the new version. Except as expressly stated in Sections
             2(a) and 2(b) above, Recipient receives no rights or
             licenses to the intellectual property of any Contributor
             under this Agreement, whether expressly, by implication,
             estoppel or otherwise. All rights in the Program not
             expressly granted under this Agreement are reserved.


             This Agreement is governed by the laws of the State of New
             York and the intellectual property laws of the United
             States of America. No party to this Agreement will bring a
             legal action under this Agreement more than one year after
             the cause of action arose. Each party waives its rights to
             a jury trial in any resulting litigation. 



*/

/* (C) COPYRIGHT International Business Machines Corp. 2001,2002          */


// File:  mech_dsa.c
//
// Mechanisms for DSA
//
// Routines contained within:

#include <pthread.h>
#include <string.h>            // for memcmp() et al
#include <stdlib.h>

#include "pkcs11types.h"
#include "defs.h"
#include "host_defs.h"
#include "h_extern.h"
//#include "tok_spec_struct.h"



//
//
CK_RV
dsa_sign( SESSION             * sess,
          CK_BBOOL              length_only,
          SIGN_VERIFY_CONTEXT * ctx,
          CK_BYTE             * in_data,
          CK_ULONG              in_data_len,
          CK_BYTE             * out_data,
          CK_ULONG            * out_data_len )
{
   OBJECT          *key_obj   = NULL;
   CK_ATTRIBUTE    *attr      = NULL;
   CK_BYTE          sig[DSA_SIGNATURE_SIZE];
   CK_OBJECT_CLASS  class;
   CK_BBOOL         flag;
   CK_RV            rc;


   rc = object_mgr_find_in_map1( ctx->key, &key_obj );
   if (rc != CKR_OK){
      st_err_log(118, __FILE__, __LINE__);
      return rc;
   }
   // must be a PRIVATE key operation
   //
   flag = template_attribute_find( key_obj->template, CKA_CLASS, &attr );
   if (flag == FALSE){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }
   else
      class = *(CK_OBJECT_CLASS *)attr->pValue;

   // if it's not a private DSA key then we have an internal failure...means
   // that somehow a public key got assigned a CKA_SIGN attribute
   //
   if (class != CKO_PRIVATE_KEY){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }

   // check input data length restrictions.  Generic DSA works on the SHA-1
   // hash of the data so the input to the DSA operation must be 20 bytes
   //
   if (in_data_len != 20){
      st_err_log(109, __FILE__, __LINE__);
      return CKR_DATA_LEN_RANGE;
   }
   if (length_only == TRUE) {
      *out_data_len = DSA_SIGNATURE_SIZE;
      return CKR_OK;
   }

   rc = ckm_dsa_sign( in_data, sig, key_obj );
   if (rc == CKR_OK) {
      memcpy( out_data, sig, DSA_SIGNATURE_SIZE );
      *out_data_len = DSA_SIGNATURE_SIZE;
   }

   return rc;
}


//
//
CK_RV
dsa_verify( SESSION             * sess,
            SIGN_VERIFY_CONTEXT * ctx,
            CK_BYTE             * in_data,
            CK_ULONG              in_data_len,
            CK_BYTE             * signature,
            CK_ULONG              sig_len )
{
   OBJECT          *key_obj  = NULL;
   CK_ATTRIBUTE    *attr     = NULL;
   CK_OBJECT_CLASS  class;
   CK_BBOOL         flag;
   CK_RV            rc;


   rc = object_mgr_find_in_map1( ctx->key, &key_obj );
   if (rc != CKR_OK){
      st_err_log(118, __FILE__, __LINE__);
      return rc;
   }
   // must be a PUBLIC key operation
   //
   flag = template_attribute_find( key_obj->template, CKA_CLASS, &attr );
   if (flag == FALSE){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }
   else
      class = *(CK_OBJECT_CLASS *)attr->pValue;

   if (class != CKO_PUBLIC_KEY){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }

   // check input data length restrictions
   //
   if (sig_len != DSA_SIGNATURE_SIZE){
      st_err_log(46, __FILE__, __LINE__);
      return CKR_SIGNATURE_LEN_RANGE;
   }
   if (in_data_len != 20){
      st_err_log(109, __FILE__, __LINE__);
      return CKR_DATA_LEN_RANGE;
   }
   rc = ckm_dsa_verify( signature, in_data, key_obj );
   if (rc != CKR_OK)
      st_err_log(121, __FILE__, __LINE__);
   return rc;
}


//
// mechanisms
//


//
//
CK_RV
ckm_dsa_key_pair_gen( TEMPLATE  * publ_tmpl,
                      TEMPLATE  * priv_tmpl )
{
   CK_ATTRIBUTE       * prime     = NULL;
   CK_ATTRIBUTE       * subprime  = NULL;
   CK_ATTRIBUTE       * base      = NULL;
   CK_ATTRIBUTE       * priv_exp  = NULL;
   CK_ATTRIBUTE       * publ_exp  = NULL;
   CK_ATTRIBUTE       * attr      = NULL;
   CK_BYTE            * ptr       = NULL;
   CK_BYTE              repl_buf[5500];
   CK_BBOOL             flag;
   CK_RV                rc;


   rc = token_specific_dsa_generate_keypair(publ_tmpl,priv_tmpl);
   if (rc != CKR_OK)
      st_err_log(91, __FILE__, __LINE__);
   return rc;
}


//
//
CK_RV
ckm_dsa_sign( CK_BYTE   * in_data,
              CK_BYTE   * signature,
              OBJECT    * priv_key )
{
   CK_ATTRIBUTE     * prime     = NULL;
   CK_ATTRIBUTE     * subprime  = NULL;
   CK_ATTRIBUTE     * base      = NULL;
   CK_ATTRIBUTE     * exponent  = NULL;
   CK_ATTRIBUTE     * attr      = NULL;
   CK_BYTE          * ptr       = NULL;
   CK_OBJECT_CLASS    keyclass;
   CK_RV              rc;

   rc = template_attribute_find( priv_key->template, CKA_CLASS, &attr );
   if (rc == FALSE){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }
   else
      keyclass = *(CK_OBJECT_CLASS *)attr->pValue;

   // this had better be a private key
   //
   if (keyclass != CKO_PRIVATE_KEY){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }

   rc = tok_dsa_sign(in_data, signature, priv_key);
   if (rc != CKR_OK)
      st_err_log(122, __FILE__, __LINE__);
   return rc;
}


//
//
CK_RV
ckm_dsa_verify( CK_BYTE   * signature,
                CK_BYTE   * data,
                OBJECT    * publ_key )
{
   CK_ATTRIBUTE     * prime     = NULL;
   CK_ATTRIBUTE     * subprime  = NULL;
   CK_ATTRIBUTE     * base      = NULL;
   CK_ATTRIBUTE     * exponent  = NULL;
   CK_ATTRIBUTE     * attr      = NULL;
   CK_BYTE          * ptr       = NULL;
   CK_OBJECT_CLASS    keyclass;
   CK_RV              rc;

   rc = template_attribute_find( publ_key->template, CKA_CLASS, &attr );
   if (rc == FALSE){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }
   else
      keyclass = *(CK_OBJECT_CLASS *)attr->pValue;

   // this had better be a private key
   //
   if (keyclass != CKO_PUBLIC_KEY){
      st_err_log(118, __FILE__, __LINE__);
      return CKR_FUNCTION_FAILED;
   }
   rc = tok_dsa_verify( signature, data, publ_key);
   if (rc != CKR_OK)
      st_err_log(121, __FILE__, __LINE__);
   return rc;
}

==> opencryptoki_3.7.0%2Bdfsg.orig.tar.gz: opencryptoki-3.7.0+dfsg/usr/lib/pkcs11/common/mech_dsa.c <==
/*
 * COPYRIGHT (c) International Business Machines Corp. 2001-2017
 *
 * This program is provided under the terms of the Common Public License,
 * version 1.0 (CPL-1.0). Any use, reproduction or distribution for this
 * software constitutes recipient's acceptance of CPL-1.0 terms which can be
 * found in the file LICENSE file or at
 * https://opensource.org/licenses/cpl1.0.php
 */

// File:  mech_dsa.c
//
// Mechanisms for DSA
//
// Routines contained within:

#include <pthread.h>
#include <string.h>            // for memcmp() et al
#include <stdlib.h>

#include "pkcs11types.h"
#include "defs.h"
#include "host_defs.h"
#include "h_extern.h"
#include "tok_spec_struct.h"
#include "trace.h"


//
//
CK_RV
dsa_sign( SESSION             * sess,
          CK_BBOOL              length_only,
          SIGN_VERIFY_CONTEXT * ctx,
          CK_BYTE             * in_data,
          CK_ULONG              in_data_len,
          CK_BYTE             * out_data,
          CK_ULONG            * out_data_len )
{
   OBJECT          *key_obj   = NULL;
   CK_ATTRIBUTE    *attr      = NULL;
   CK_BYTE          sig[DSA_SIGNATURE_SIZE];
   CK_OBJECT_CLASS  class;
   CK_BBOOL         flag;
   CK_RV            rc;


   rc = object_mgr_find_in_map1( ctx->key, &key_obj );
   if (rc != CKR_OK) {
      TRACE_ERROR("Failed to acquire key from specified handle");
      if (rc == CKR_OBJECT_HANDLE_INVALID)
          return CKR_KEY_HANDLE_INVALID;
      else
        return rc
   }

   // must be a PRIVATE key operation
   //
   flag = template_attribute_find( key_obj->template, CKA_CLASS, &attr );
   if (flag == FALSE){
      TRACE_ERROR("Could not find <the_attribute_name> in the template\n");
      return CKR_FUNCTION_FAILED;
   }
   else
      class = *(CK_OBJECT_CLASS *)attr->pValue;

   // if it's not a private DSA key then we have an internal failure...means
   // that somehow a public key got assigned a CKA_SIGN attribute
   //
   if (class != CKO_PRIVATE_KEY){
      TRACE_ERROR("This operation requires a private key.\n");
      return CKR_KEY_FUNCTION_NOT_PERMITTED;
   }

   // check input data length restrictions.  Generic DSA works on the SHA-1
   // hash of the data so the input to the DSA operation must be 20 bytes
   //
   if (in_data_len != 20){
      TRACE_ERROR("%s\n", ock_err(ERR_DATA_LEN_RANGE));
      return CKR_DATA_LEN_RANGE;
   }
   if (length_only == TRUE) {
      *out_data_len = DSA_SIGNATURE_SIZE;
      return CKR_OK;
   }

   rc = ckm_dsa_sign( in_data, sig, key_obj );
   if (rc == CKR_OK) {
      memcpy( out_data, sig, DSA_SIGNATURE_SIZE );
      *out_data_len = DSA_SIGNATURE_SIZE;
   }

   return rc;
}


//
//
CK_RV
dsa_verify( SESSION             * sess,
            SIGN_VERIFY_CONTEXT * ctx,
            CK_BYTE             * in_data,
            CK_ULONG              in_data_len,
            CK_BYTE             * signature,
            CK_ULONG              sig_len )
{
   OBJECT          *key_obj  = NULL;
   CK_ATTRIBUTE    *attr     = NULL;
   CK_OBJECT_CLASS  class;
   CK_BBOOL         flag;
   CK_RV            rc;


   rc = object_mgr_find_in_map1( ctx->key, &key_obj );
   if (rc != CKR_OK) {
      TRACE_ERROR("Failed to acquire key from specified handle");
      if (rc == CKR_OBJECT_HANDLE_INVALID)
          return CKR_KEY_HANDLE_INVALID;
      else
        return rc
   }

   // must be a PUBLIC key operation
   //
   flag = template_attribute_find( key_obj->template, CKA_CLASS, &attr );
   if (flag == FALSE){
      TRACE_ERROR("Could not find CKA_CLASS in the template\n");
      return CKR_FUNCTION_FAILED;
   }
   else
      class = *(CK_OBJECT_CLASS *)attr->pValue;

   if (class != CKO_PUBLIC_KEY){
      TRACE_ERROR("This operation requires a public key.\n");
      return CKR_KEY_FUNCTION_NOT_PERMITTED;
   }

   // check input data length restrictions
   //
   if (sig_len != DSA_SIGNATURE_SIZE){
      TRACE_ERROR("%s\n", ock_err(ERR_SIGNATURE_LEN_RANGE));
      return CKR_SIGNATURE_LEN_RANGE;
   }
   if (in_data_len != 20){
      TRACE_ERROR("%s\n", ock_err(ERR_DATA_LEN_RANGE));
      return CKR_DATA_LEN_RANGE;
   }
   rc = ckm_dsa_verify( signature, in_data, key_obj );
   return rc;
}


//
// mechanisms
//


//
//
CK_RV
ckm_dsa_key_pair_gen( TEMPLATE  * publ_tmpl,
                      TEMPLATE  * priv_tmpl )
{
   CK_ATTRIBUTE       * prime     = NULL;
   CK_ATTRIBUTE       * subprime  = NULL;
   CK_ATTRIBUTE       * base      = NULL;
   CK_ATTRIBUTE       * priv_exp  = NULL;
   CK_ATTRIBUTE       * publ_exp  = NULL;
   CK_ATTRIBUTE       * attr      = NULL;
   CK_BYTE            * ptr       = NULL;
   CK_BYTE              repl_buf[5500];
   CK_BBOOL             flag;
   CK_RV                rc;


   if (token_specific.t_dsa_generate_keypair == NULL) {
      TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
      return CKR_MECHANISM_INVALID;
   }

   rc = token_specific.t_dsa_generate_keypair(publ_tmpl,priv_tmpl);
   if (rc != CKR_OK)
      TRACE_DEVEL("Tpken specific dsa keypair generation failed.\n");
   return rc;
}


//
//
CK_RV
ckm_dsa_sign( CK_BYTE   * in_data,
              CK_BYTE   * signature,
              OBJECT    * priv_key )
{
   CK_ATTRIBUTE     * prime     = NULL;
   CK_ATTRIBUTE     * subprime  = NULL;
   CK_ATTRIBUTE     * base      = NULL;
   CK_ATTRIBUTE     * exponent  = NULL;
   CK_ATTRIBUTE     * attr      = NULL;
   CK_BYTE          * ptr       = NULL;
   CK_OBJECT_CLASS    keyclass;
   CK_RV              rc;

   rc = template_attribute_find( priv_key->template, CKA_CLASS, &attr );
   if (rc == FALSE){
      TRACE_ERROR("Could not find CKA_CLASS in the template\n");
      return CKR_FUNCTION_FAILED;
   }
   else
      keyclass = *(CK_OBJECT_CLASS *)attr->pValue;

   // this had better be a private key
   //
   if (keyclass != CKO_PRIVATE_KEY){
      TRACE_ERROR("This operation requires a private key.\n");
      return CKR_KEY_FUNCTION_NOT_PERMITTED;
   }

   if (token_specific.t_dsa_sign == NULL) {
      TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
      return CKR_MECHANISM_INVALID;
   }

   rc = token_specific.t_dsa_sign(in_data, signature, priv_key);
   if (rc != CKR_OK)
      TRACE_DEVEL("Token specific dsa sign failed.\n");
   return rc;
}


//
//
CK_RV
ckm_dsa_verify( CK_BYTE   * signature,
                CK_BYTE   * data,
                OBJECT    * publ_key )
{
   CK_ATTRIBUTE     * prime     = NULL;
   CK_ATTRIBUTE     * subprime  = NULL;
   CK_ATTRIBUTE     * base      = NULL;
   CK_ATTRIBUTE     * exponent  = NULL;
   CK_ATTRIBUTE     * attr      = NULL;
   CK_BYTE          * ptr       = NULL;
   CK_OBJECT_CLASS    keyclass;
   CK_RV              rc;

   rc = template_attribute_find( publ_key->template, CKA_CLASS, &attr );
   if (rc == FALSE){
      TRACE_ERROR("Could not find CKA_CLASS in the template\n");
      return CKR_FUNCTION_FAILED;
   }
   else
      keyclass = *(CK_OBJECT_CLASS *)attr->pValue;

   // this had better be a private key
   //
   if (keyclass != CKO_PUBLIC_KEY){
      TRACE_ERROR("This operation requires a public key.\n");
      return CKR_KEY_FUNCTION_NOT_PERMITTED;
   }

   if (token_specific.t_dsa_verify == NULL) {
      TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
      return CKR_MECHANISM_INVALID;
   }
   rc = token_specific.t_dsa_verify(signature, data, publ_key);
   if (rc != CKR_OK)
      TRACE_DEVEL("Token specific dsa verify failed.\n");
   return rc;
}
